// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
// 																	hmm.lrl: left-right-left Hidden Markov model for 
// 														 							 international event sequences
//__________________________________________________________________________________
// 
//
// ORIGINAL COPYRIGHT INFORMATION
//   file : hmm.cc
// version: 1.03 [August 21, 1995]
/* Copyright (C) 1994 Richard Myers and James Whitson
   Permission is granted to any individual or institution to use, copy, or
   redistribute this software so long as all of the original files are 
   included unmodified, that it is not sold for profit, and that this 
   copyright notice is retained.
*/

/*  *************************************************************
    Updated by Emdad Khan (Computer & Information Science dept,
    U.C Santa Cruz and National Semiconductor, Santa Clara, CA).
    Date 5/12/95. Brief description of the changes follows:

    (1)
    The original model written by Richard & James seemed to works for equal
    length string only. For unequal length strings (i.e vq coded words),
    i.e when different people say the same words, the program does not 
    converge (basically oscillates). The authors used "instanteneous"
    updates for each strings for the unequal case. Such scheme nullifies
    the learning a string by another etc. The commonly used approach
    is the "batch" update: a good reference is Rabiner's book or paper
    (as alos referenced by the authors). 
     
    So, I updated the code (training routine and other associated routines)
    to use batch update for unequal length strings. More detailed descrip-
    tion of the changes are given in corresponding routines i.e
    batch_train, beta_I, compute_gamma, alpha_F and reestimate routines.

    (2) 
    Some of equations related to training (namely, gamma, beta, alpha)
    showed discrepencies compared to the equations given in HMM literatures.
    Those equations are corrected and details of the changes are shown in
    corresponding routines. These routines are basically same as the 
    routines mentioned above.

    (3) 
    Some of the parameters were initialised both in the main routines as 
    well as in the function declarations in the .h file. Some of the 
    compilers does not like this. So, initialization of parameters are
    changed so that it is present only in one place. 
    
    *******************************************************************
*/ 
 
/*  *************************************************************
    Updated by Arthur Stephens (Arthur.Stephens@src.bae.co.uk)
    Date 8/30/95.

    Purified code, plugged memory leaks, corrected some out of bounds
    array conditions and got it to compile on Solaris. (Thanks Art --rm)
*/

/*  *************************************************************
    Modifications by Philip A. Schrodt (p-schrodt@ukans.edu)
    Date: 97.3.10.

    1. Files were translated into Metrowerks C (Macintosh).  This involved moving
       from the object model of C++ to a bunch of clunky globals and combining
       most of the original files.  Less elegant but it works...
       
    Modifications for hmm.pa.
    Date: 97.6.22.
    
    1. Code for doing the tests against BCOW data removed.
    
    2. N_EXPER different initial values were used and the best MAX_SAVE of these
       are saved and written to a file.
       
    3. Nonwar/war tests were automated
    
    Modifications for hmm.lrl.
    Date: 97.7/8.

    1. Added code to implement the Viterbi algorithm for generating the most
       probable state sequence (from Rabiner 1989) 
       
    2. Added a transition to previous state; eliminated the transition symbol
    	 vectors.  fCIRC controls whether chain is circular.   Rewrote the estimation 
    	 routines to match Rabiner.   I/O formats changed to handle the LRL parameters.
    	 
    	 Also added the fLR switch that forces the model to be LR, albeit with the
    	 new estimation code rather than M&W's.
    	 
    3. Eliminated the equal-length case code;
    
    4. Added a system for estimating cases based on the SEQ_SIZE events prior to a 
       set of dates read from a file.
       
    5. Added an Count_TFT as an independent utility for counting cases of TFT 
    	 behavior.
    	 
    6. Added system for handling dyadic behaviors by concatenating the obverse
       dyads code with an increment of 22 to the original dyad; most of the code
       and files for this involve a DY suffix somewhere.
cc hmm.casia.pr.c -lmpi -lm -Xcpluscomm -woff all
    	 
    7. Added another Monte-Carlo layer to look for a maximum involving the cross-
       correlation ("xcorr") of the percentage of time in various state probabilities
       and an external indicator (e.g. TFT measure or Goldstein; this is read from
       a file.  Distribution of the xcorrs is also recorded.
       
    8. Continued to control all of these options with a rat's-nest of #if statements... 
    
    9. Added matrices to estimate the mean and stdev of the parameters 
         
*/

#define	FALSE							0
#define TRUE							1

//_________________ M O D E L  S P E C I F I C A T I O N ___________________

#define FOUR_WEEK					// 4 week window
//#define EIGHT_WEEK					// 8 week window

//#define U_MODEL					// unweighted model
//#define P_MODEL					// entropy weights; high weight on false positive
#define N_MODEL						// entropy weights; high weight on false negative

//#define FORECAST				 184	// number of days for forecast -- 6 month
#define FORECAST					91	// number of days for forecast -- 3 month
//#define FORECAST					28	// number of days for forecast -- 1 month

//#define AFG					// choose the dyad combinations
//#define ARM
//#define TAJ

#define LIBR

// Parameters set in this block:
//  MAX_SYMBOLS			 	Number of symbols in the HMM model; this is (N_CODES * N_DYADS) + 1
//                   	where N_CODES = # of event codes (4 in this configuration)
//                         N_DYADS = # of dyads
//                   	(the +1 comes from the non-event code)
//                   	The N_DYADS isn't used explicitly in this program; it is implicit in
//                   	the creation of the SEQUENCE_FILE.
//	SEQUENCE_FILE			File containing the sequence data
//	WEIS_COUNT_FILE 	File containing the number of violent events per week
//	RESULTS_FILE			Output file name template -- various pieces are replaced in Init_File(int rank)
// 	CONFL_THRESHOLD  	Number of violent events required for a "high conflict" week


#ifdef  LIBR				     	// 8-dyad analysis for Afghanistan
	#define MAX_SYMBOLS			 33		// transformed 4-code WEIS sequences
	#define	SEQUENCE_FILE		"LIBR.SEQ"
	#ifdef FOUR_WEEK
		#define	WEIS_COUNT_FILE "LIBR.4.COUNT"
		#define	RESULTS_FILE		"FS.LIBR.4UL.RR.XXXX"
		#define CONFL_THRESHOLD  2.5    // threshold for 4-week window
	#else
		#define	WEIS_COUNT_FILE "LIBR.8.COUNT"
		#define	RESULTS_FILE		"FS.LIBR.8UL.RR.XXXX"
		#define CONFL_THRESHOLD  4.5    // threshold for 8-week window
	#endif
	#define BASE_DATE       31272   // first day of sequence data to use
	#define FIRST_YEAR				1989	// first year of the prediction analysis
	#define FIRST_YEAR_INDEX	1			// first index of the prediction loop
#endif

#ifdef  AFG				     	// 6-dyad analysis for Afghanistan
	#define MAX_SYMBOLS			 25		// transformed 4-code WEIS sequences
	#define	SEQUENCE_FILE		"AFG.ALL.SQ"
	#ifdef FOUR_WEEK
		#define	WEIS_COUNT_FILE "AFG.4.COUNT"
		#define	RESULTS_FILE		"PR.AFGR.4UL.RR.XXXX"
		#define CONFL_THRESHOLD  19.5    // threshold for 4-week window
	#else
		#define	WEIS_COUNT_FILE "AFG.8.COUNT"
		#define	RESULTS_FILE		"PR.AFGR.8UL.RR.XXXX"
		#define CONFL_THRESHOLD  39.5    // threshold for 8-week window
	#endif
	#define BASE_DATE       32142    // first day of sequence data to use
	#define FIRST_YEAR				1989	// first year of the prediction analysis
	#define FIRST_YEAR_INDEX	4			// first index of the prediction loop
#endif

#ifdef  ARM				   // 2-dyad analysis for Armenia-Azerbijan
	#define MAX_SYMBOLS			 9
	#define	SEQUENCE_FILE		"ARMAZR.ALL.SQ"
	#ifdef FOUR_WEEK
		#define	WEIS_COUNT_FILE "ARMAZR.4.COUNT"
		#define	RESULTS_FILE		"PR.ARAZ.4UL.RR.XXXX"
		#define CONFL_THRESHOLD  5.5     // threshold for 4-week window
	#else
		#define	WEIS_COUNT_FILE "ARMAZR.8.COUNT"
		#define	RESULTS_FILE		"PR.ARAZ.8UL.RR.XXXX"
		#define CONFL_THRESHOLD  13.5    // threshold for 8-week window
	#endif
	#define BASE_DATE       32142    // first day of sequence data to use
	#define FIRST_YEAR				1989	// first year of the prediction analysis
	#define FIRST_YEAR_INDEX	4			// first index of the prediction loop
#endif										// end dyad specification

#ifdef  TAJ					// 2-dyad analysis for Tajikistan-Russia
	#define MAX_SYMBOLS			9		// transformed WEIS sequences, Balkans
	#define	SEQUENCE_FILE		"RUSTAJ.ALL.SQ"
	#ifdef FOUR_WEEK
		#define	WEIS_COUNT_FILE "RUSTAJ.4.COUNT"
		#define	RESULTS_FILE		"PR.RUTA.4UL.RR.XXXX"
		#define CONFL_THRESHOLD  3.5     // threshold for 4-week window
	#else
		#define	WEIS_COUNT_FILE "RUSTAJ.8.COUNT"
		#define	RESULTS_FILE		"PR.RUTA.8UL.RR.XXXX"
		#define CONFL_THRESHOLD  7.5     // threshold for 8-week window
	#endif
	#define BASE_DATE       32142    // first day of sequence data to use
	#define FIRST_YEAR				1989	// first year of the prediction analysis
	#define FIRST_YEAR_INDEX	4			// first index of the prediction loop
#endif


#define USE_MPI					// enable MPI code

#define	N_TEMPLATES			8        // number of template used to estimate model

//___________________________________________________________________________


#define MAX_LENGTH		  128		// max length of training strings
#define MAX_TRELLIS		  MAX_LENGTH+1		// determined by maximum sequence length
#define MAX_STATES			  6
#define MAX_STRINGS			 N_TEMPLATES		// maximum number of training strings 
#define	NAME_LENGTH			 32
#define	MAX_SAVE					4		// number of experiments saved
#define	MIN_DELTA				1.0		// min_delta_psum  (originally 0.01)
#define	MAX_WEIS			40000		// maximum aWEIS size
#define SEQ_SIZE 				100		// prior sequence length (actual training string length)
#define MAX_MONTHS 			240		// max length of monthly data in indic. file
#define FITNESS_BASE	  100.0	// offset to insure all fitness > 0 in XC_GA

#define	N_EXPER				   32		// number of experiments with random initial matrices

															// values for the globals max_symbols and max_states
#define RUN_SYMBOLS			 MAX_SYMBOLS
#define RUN_STATES			 MAX_STATES

															// GA parameters
#define	GA_VECT					32		// number of vectors used
#define	GA_SURVIVE			 0.25	// survival rate
#define	GA_MUTATE				 0.1	// mutation rate
#define	GA_MAX_EON		  32		// max eons for GA
#define	GA_EXPER				16		// number of GA experiments

#define MAX_WEEKS			 	 512	// max number of weeks of data
#define MAX_TEMPLATES		N_TEMPLATES	// number of random templates used

#define N_YEARS			 			 11	// number of years indexed in the data (1989 - 1999)

// flags
#define fCIRC			 		FALSE		// circular matrix?
#define fLR						FALSE		// LR matrix?
#define fWRITE_ALPHA	FALSE		// write the alphas during estimation?
#define fWRITE_FILE		FALSE		// write a file of the monthly alphas?  Set FALSE when doing an XC estimation
#define XC_GA					FALSE		// optimize for XC using the GA


//_____________________________________________________________________________
// Typedefs and structs

typedef double arTrellState[MAX_TRELLIS][MAX_STATES] ;	
typedef double arTrell[MAX_TRELLIS] ;	
typedef double arState[MAX_STATES] ;	
typedef double arStateSymb[MAX_STATES][MAX_SYMBOLS] ;	
typedef double arNumSymb[MAX_SYMBOLS];	
typedef double arSymb[MAX_SYMBOLS];	

typedef struct 
{ arNumSymb 	symb_out; 				// array of weights for output symbols
  double 			prev_trans;
  double 			recur_trans;
  double 			next_trans;
} state; 

struct hist		// for aHistory
{ double		val;
	int				age, top;};

struct count22		// for aCount22
{ long		date;
	int			count;};

//___________________________________________________________________________
// Prototypes

//	------- Meyers and Whitson main programs -------
void Train_HMM(int seq_length);

//	------- Utilities -------
void Initialize (int rank);
void Calc_Memory( void );					// computes memory required in main arrays
double UnifRnd( void );						// uniform [0,1] random number generator

//	------- aSTATE FUNCTIONS -------
double set_symb_out(state *aState,int symbol, double prob);
double set_recur_trans(state *aState,double prob);
double set_next_trans(state *aState, double prob);
double set_prev_trans(state *aState, double prob);

//	------- MODEL INIT -------
void rnd_init_hmm_probs( void );
void rnd_sym_probs(double prob_syms[]);

//	------ CALCULATIONS -------
void rescale_alphas(int col);
void rescale_betas(int col);
double alpha_F(int* symbol_array, int seq_length);
double beta_I(int* symbol_array, int seq_length);
void compute_gamxi(int* symbol_array, int seq_length);
double a_numer(int i, int j, int seq_length);						// also b_denom
double a_denom(int i, int seq_length);
double b_numer(int i, int sym, int *symbol_array, int seq_length);
double b_denom(int i, int seq_length);
double set_ab_counts(int* string_array,int seq_length);
double reestimate(void);																		// returns sum of changes
double test(int* string, int seq_length);										// test one string 

//	------ VITERBI ROUTINES -------

void Viterbi( int index);									// compute MLE state sequence for string[index] using	Viterbi algorithm from Rabiner
double qprop(int state, int length);			// return proportion of qstar that have the value state

 //	------ HMM ROUTINES -------

void Set_Size(void);																			// set globals max_symbols and max_states
void Init_Random_HMM(void);																// initialize random model
void batch_train(int seq_length, double min_delta_psum);	// estimate model
void check_save(void);																		// check whether we should save the model
void show_probs(void);                         						// display model probabilities
void dump_model(char *fname);           									// dumps the current model

//	------ WEIS ROUTINES -------

long Julian (char *s);																		// Converts dates of form YYMMDD to Julian serial beginning 1 Jan 1900 
void Read_WEISSQ(char *sfile);																// read a dyadic WEISSQ file
void Estimate_HMM(void);																	// MC estimation from existing set of strings
void Next_Month (char *sMonth);														// increment the month string
void Compute_Running_Prob (char *SourceCode, char *TargetCode, char ctype );			// computes alpha for 100 events prior to the end of each month

//	------ GA ROUTINES -------

void Evolve(int eon);												// does genetic selection, recombination and mutation on existing vectors
void Mutate_TransProbs(int kidx, int ka);	//	generate new transition probabilites for an LRL model
void Initialize_Vectors(void);						// set up initial vectors
void Estimate_GA(void);										// Main GA estimation loop
void Estimate_Vector(int idx);						// Estimation starting from aGAStart[idx]
void init_GA_probs(int idx);							// initialize the model parameters from aGAStart[idx]
void Init_Saved_HMMGA(int top);						// set max_states, max_symbols and model parameters from file
void Write_GA_History(int eon);						// write the aHistory array to screen
void File_Results(int nex);								// writes the results to fres at the end of evolution
void File_Model(void);										// write the model in aState to fres
void Create_File_Name(void);				// Creates a unique output file name for fres
void Init_File(int rank);						// initializes fres with rank id


//	------ BALKANS/CASIA MODEL ROUTINES -------

void Read_Count(void);										// read the Count22 file -- number of WEIS 22 events by week
long Rand_HighConflict(void);							// returns a randomly-chosen high conflict week
long Rand_LowConflict(void);							// returns a randomly-chosen low conflict week
void Get_Template_Dates(void);						//
int Verify_Sequence_Start(long ijul);			// Finds the aWEIS index of the beginning of a SEQ_SIZE sequence that ends on ijul
void Init_Template_HMM(long *aDates);			// Reads a set of dates from aDates minus FORECAST; moves SEQ_SIZE events into strings, 
void Compute_Weekly_Prob (float *probs );	// computes alpha for SEQ_SIZE events prior to the beginning of a week, minus
void Write_Forecast (int isFull);					// write various summary results to fres; TRUE is all weeks; other predicted weeks
float Compare_High_Low(void);							// calculates comparison statistic for aHighProb[ka] - aLowProb[ka]


//__________________________________________________________
// Global variables
  
const double MIN_PROB = 0.000001;	// never let alpha or beta equal zero

// Monte-Carlo experiments
state			aStates[MAX_STATES];   						// matrix of model parameters
state			aSaveSt[MAX_STATES][MAX_SAVE];   	// array for saving models
double		aSaveAlpha[MAX_SAVE];   					// sum of alpha of saved models
double		curAlpha;													// tot_alpha from last computation
int				string_zero[MAX_LENGTH];   				// copy of string[0], which is overwritten in the XC test

state			aGAFinal[MAX_STATES][2*GA_VECT];  	// array for GA models final estimates
state			aGAStart[MAX_STATES][2*GA_VECT];  	// array for GA models starting points
double		aGAFit[GA_VECT];   								// GA fitness
int				aGAAge[GA_VECT];   								// GA age
int				eon;															// current eon
struct hist	aHistory[GA_MAX_EON];
 
state			aSumState[MAX_STATES];   				// sum of states; used to estimate variance
state			aSqrState[MAX_STATES];   				// square of states

struct count22	aCount22[MAX_WEEKS];				// holds count of 22 events by week
int				N_Weeks;
long			aHigh[MAX_TEMPLATES];						// dates of templates
long			aLow[MAX_TEMPLATES];
float			aHighProb[MAX_WEEKS],aLowProb[MAX_WEEKS];	// alphas for high, low models

// from class HMM:

int		max_states;                   			// number of states in the model
int		max_symbols;                  			// symbol emissions per state
int		trellis_width;                			// cols allocated in trellis =num_symbols-1
int		num_strings;           							// number of training/testing strings
int		strings[MAX_STRINGS][MAX_LENGTH];   // matrix of training/testing strings
int		string_len[MAX_STRINGS];		    		// sequence length
char	string_name[MAX_STRINGS][NAME_LENGTH];		// sequence name
double					prob_seq;								// sequence probability given current model (computed in alpha_F)

arTrellState 		alpha;                  // matrix for alpha trellis
arTrellState 		beta;                   // matrix for beta trellis
arTrellState 		gammaF;                  // matrix for gamma trellis
arTrell				  scaling_factors;        // array of alpha and beta scaling factors
arTrellState 		xi_prev;             		// matrix of xi prev transition probs
arTrellState 		xi_next;             		// matrix of xi next transition probs
arTrellState 		xi_recur;            		// matrix of xi recurrence probs
arState 				a_numer_sum_recur;      // array of numerators for a_ij's
arState 				a_numer_sum_next;       // array of numerators for a_ij's
arState 				a_numer_sum_prev;       // array of numerators for a_ij's
arState 				a_denom_sum;      			// array of denomonators for a_ij's
arStateSymb 		b_numer_sum;      // array of numerators for b_ij's
arState 				b_denom_sum;       // array of denominators for b_ij's

arTrellState 		delta;                  // matrix for delta trellis (Viterbi alg.)
int 						psi[MAX_TRELLIS][MAX_STATES] ; // same for gamma	
int 						qstar[MAX_TRELLIS] ;		// holds the optimal state sequence

long						aWEIS_Date[MAX_WEIS];		// dates of WEIS events
int							aWEIS_Evt[MAX_WEIS];		// WEIS event sequence
int							NumWEIS;								// number of WEIS events
int			Month[] = {0,0,31,59,90,120,151,181,212,243,273,304,334};		// used for Julian computation: # days prior to Month[n] JAN=1
FILE						*fres;									// holds results of NEX experiment loop 

// variables added for the prediction analysis

int		Year;															// year index for start of prediction period
int		N_High[N_YEARS], N_Low[N_YEARS]; // number of high, low conflict weeks by year
int 	Count_Year[N_YEARS];							// aCount index for start of each year
int		YearStart[] = {31047, 31412, 31777, 32142, 32508, 32873, 33238, 33603, 33969, 34334, 34699}; // Julian dates for 1 January in 1992 through 2000
int		fFullForecast = FALSE;						// compute alpha probs for all weeks; otherwise just prior period
int		fNoTemplates = FALSE;						// error condition when insufficient templates exist
